#!/bin/bash
# 'set -f' disables filename expansion (globbing)
# This should be safer in preventing variales like packageFiles from expansion
set -f
PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin
# Repo API Service parameters
masterPort=22
secondaryPort=22202
hostname=azure-apt-cat.cloudapp.net
sshUser=azure-aptcatalog

aptcatalogHome=/var/lib/azure-aptcatalog
aptRemoteRoot=$aptcatalogHome/repos
aptRemoteAccounts=/etc/azure-aptcatalog/repo_accounts.csv
aptUnlockFile=$aptcatalogHome/unlocked
yumRemoteRoot=$aptcatalogHome/yumrepos
yumRemoteAccounts=/etc/azure-aptcatalog/yum_accounts.csv
yumUnlockFile=$aptcatalogHome/yum_unlocked
wwwDir=/var/www/html/repos
wwwUser=www-data
open="open"

# Local Parameters
accounts=/etc/azuremirror/repo_accounts.csv
tmpAccounts=$accounts.tmp
reporoot=/var/reporoot
sshdConfig=/etc/ssh/sshd_config
sshKeyFile=/root/.ssh/reposync_rsa
statsdClient="127.0.0.1:8125"

newAccts=0
runningUser=$USER
syslogFacility=syslog
syslogTag=SyncMirror
exec > /var/log/reposync.log 2>&1

# Function Definitions
function LogMessage {
    msg=$1
    priority=$2
    if [ -z "$msg" ];then
        return
    fi

    if [ -z "$priority" ]; then
        priority=info
    fi

    logger -p "$syslogFacility.$priority" -t $SyncMirror -- $msg
    echo "[$priority]: $msg"
}

function Bail {
    msg=$1
    if [ -z "$msg" ];then
        msg="An unspecified error occurred"
    fi
    LogMessage "$msg" error
    exit 1
}

function TestUnlockFile {
    port=$1
    user=$2
    file=$3
    scp -P $port -i $sshKeyFile $user@$hostname:$file /tmp 2> /dev/null
    return $?
}

function SetupRepoAccounts {
    sshPort=$1
    sshUser=$2
    accountFile=$3

    # Get CSV file
    install -o root -g root -m 600 /dev/null $tmpAccounts
    scp -P $sshPort -i $sshKeyFile $sshUser@$hostname:$accountFile $tmpAccounts

    if [[ $? -ne 0 || ! -f $tmpAccounts ]]; then
        Bail "File $tmpAccounts was not successfully copied from master"
    fi

    # Move CSV file to a location where users cannot meddle with it
    mv $tmpAccounts $accounts
    if [ -f $accounts ]; then
        LogMessage "Successfully retrieved $accounts"
    else
        Bail "Failed to move $tmpAccounts to $accounts"
    fi

    # Create /var/www/html/repos folder if it doesn't exist
    if [ ! -f "$wwwDir" ]; then
        install -o $wwwUser -g $wwwUser -m 755 -d $wwwDir
    fi

    while IFS=, read username visibility groups folder sshkey
    do
        if [ -z "$username" ];then
            continue
        fi
        if [ "$username" == "$wwwUser" ] || [ "$visibility" == "$open" ]; then
            # Create symlink
            if [ ! -e $wwwDir/$folder ]; then
                LogMessage "Linking $wwwDir/$folder -> $reporoot/repos/$folder"
                ln -s $reporoot/repos/$folder $wwwDir/$folder
            fi
            continue
        fi
        acctExists=$(grep -c "^$username:" /etc/passwd)
        # Setup account if it doesn't already exist
        if [ $acctExists -lt 1 ]; then
            # Set up account
            LogMessage "Account $username doesn't exist. Creating it."
            useradd -d /home/$username -m -s /bin/false $username
            if [ $? -ne 0 ]; then
                LogMessage "Failed to create account for $username" warning
                continue
            fi
            install -o $username -g $username -m 700 -d /home/$username/.ssh
        fi
        if ! grep -q "^Match User $username\$" $sshdConfig; then
            LogMessage "Adding SFTP config for $username to $sshdConfig"
            # Set up SFTP permissions
            cat << EOF >> $sshdConfig

Match User $username
    ChrootDirectory $reporoot
    AllowTCPForwarding no
    X11Forwarding no
    ForceCommand internal-sftp

EOF
            let "newAccts += 1"
        fi
        # Set up ssh key Create file with appropriate permissions first
        # Then fill in the appropriate keys, then overwrite the old keys file
        authKeyFile=/home/$username/.ssh/authorized_keys
        install -o $username -g $username -m 600 /dev/null $authKeyFile.tmp
        echo "ssh-rsa $sshkey" > $authKeyFile.tmp
        mv $authKeyFile.tmp  $authKeyFile

        # Update group membership
        groupsCmd=$(echo $groups | tr ':' ',')
        LogMessage "Setting group membership for $username to [$groupsCmd]"
        usermod -G "$groupsCmd" $username
    done < $accounts

    if [ $newAccts -gt 0 ]; then
        LogMessage "Added $newAccts accounts, must restart SSH daemon"
        service ssh restart
        if [ $? -ne 0 ]; then
            Bail "Error while restarting SSH daemon"
        fi
        LogMessage "Successfully restarted SSH daemon"
    fi
}

function SyncRepoFiles {
    sshPort=$1
    sshUser=$2
    remoteRoot=$3
    packageFiles=$4

    # If rsync is already running, don't start another instance
    if ps waux | grep -v grep | grep -q rsync; then
        echo "Rsync is already running"
        exit 0
    fi
    # Rsync package files
    startTime=$(date +%s)

    # First sync all .deb files with size changed
    temp_file=$(mktemp)
    rsync -agpz --stats --delete --size-only --include="*/" --include="$packageFiles" --exclude="*" -e "ssh -p $sshPort -i $sshKeyFile" $sshUser@$hostname:$remoteRoot $reporoot > $temp_file
    errCode=$?
    total_bytes_transferred=$(cat $temp_file | grep 'Total bytes received' | sed 's/[^0-9]*//g')
    rm $temp_file
    endTime=$(date +%s)
    elapsed=$((endTime-startTime))

    # Logging time taken for rsync to statsd. We extract the VM's region by relying on our naming convention.
    LogToStatsd 'RsyncTimeInSeconds' $elapsed

    # Logging bytes transferred during the rsync to statsd.
    LogToStatsd 'RsyncBytesSent' $total_bytes_transferred

    if [ $errCode -ne 0 ];then
        Bail "Rsync $packageFiles files failed after $elapsed seconds"
    else
        LogMessage "Rsync $packageFiles files succeeded after $elapsed seconds"
    fi

    # Then sync the rest of files by comparing checksum
    temp_file2=$(mktemp)
    rsync -agpz --stats --delete --checksum --exclude="$packageFiles" -e "ssh -p $sshPort -i $sshKeyFile" $sshUser@$hostname:$remoteRoot $reporoot > $temp_file2
    errCode=$?
    total_bytes_transferred=$(cat $temp_file2 | grep 'Total bytes received' | sed 's/[^0-9]*//g')
    rm $temp_file2
    endTime=$(date +%s)
    elapsed=$((endTime-startTime))

    LogToStatsd 'RsyncTimeInSeconds' $elapsed
    LogToStatsd 'RsyncBytesSent' $total_bytes_transferred

    if [ $errCode -ne 0 ];then
        Bail "Rsync meta data failed after $elapsed seconds"
    else
        LogMessage "Rsync meta data succeeded after $elapsed seconds"
    fi
}

function LogToStatsd {
  metric_name=$1
  metric_value=$2
  statsd_counter_type='g' # This represents the Statsd counter type 'gauge'.
  region=$(echo $HOSTNAME | cut -d '-' -f 2)
  command -v socat > /dev/null 2>&1 || { LogMessage "The syncrepos.sh script requires socat to log telemetry data to Statsd. Please install socat."; return 1; }
  echo "{\"Namespace\":\"MicrosoftRepo\", \"Metric\":\"${metric_name}\", \"Dims\": {\"Region\":\"$region\", \"MachineName\":\"$HOSTNAME\"}}:${metric_value}|${statsd_counter_type}" | socat -t 1 - UDP:$statsdClient
}

function SyncRepos {
  repoType=$1

  if [ "$repoType" = "apt" ]; then
    unlockFile=$aptUnlockFile
    accountFile=$aptRemoteAccounts
    remoteRoot=$aptRemoteRoot
    pkgFiles='*.deb'
  elif [ "$repoType" = "yum" ]; then
    unlockFile=$yumUnlockFile
    accountFile=$yumRemoteAccounts
    remoteRoot=$yumRemoteRoot
    pkgFiles='*.rpm'
  else
    Bail "Unsupported repo type: $repoType"
  fi

  if TestUnlockFile $masterPort $sshUser $unlockFile; then
      # Try the primary
      LogMessage "Syncing $repoType repos from the primary repo service..."
      if [ "$repoType" = "apt" ]; then
          SetupRepoAccounts $masterPort $sshUser $accountFile
      fi
      SyncRepoFiles $masterPort $sshUser $remoteRoot $pkgFiles
  elif TestUnlockFile $secondaryPort $sshUser $unlockFile; then
      # Try the secondary
      LogMessage "Syncing $repoType repos from the secondary repo service..."
      if [ "$repoType" = "apt" ]; then
          SetupRepoAccounts $secondaryPort $sshUser $accountFile
      fi
      SyncRepoFiles $secondaryPort $sshUser $remoteRoot $pkgFiles
  else
      # If build operation is occurring, don't synchronize, could result in an incomplete copy
      LogMessage "Cannot sync $repoType repos from $hostname because its repo is locked. Waiting for the following file to be created: $hostname:$unlockFile" "warning"
  fi
}

# Pre-execution check
runningUser=$SUDO_USER
if [[ -z $SUDO_USER && "$USER" -ne "root" ]];then
    Bail "Program must run under sudo or root"
fi

SyncRepos "apt"
SyncRepos "yum"
